"""
Utility classes for network routing
"""

from seine.address import IPv4Address, IPv6Address, EthernetMACAddress
from subprocess import check_output, CalledProcessError

class RoutingTableError(Exception):
    pass

class IPRoutingTable(object):
    def __init__(self):
        self.ipv4 = []
        self.ipv6 = []

        try:
            output = check_output(['netstat', '-rn'])
        except CalledProcessError, emsg:
            raise RoutingTableError('Error checking netstat -rn output: %s' % emsg)

        for l in [x.strip() for x in output.split('\n') if x!='']:

            fields = l.split()
            if fields[0] == 'default':
                address = 'default'
            else:
                address = self.__parse_address__(fields[0])
                if address is None:
                    continue

            gateway = self.__parse_address__(fields[1])
            if gateway is not None:
                gateway_type = 'route'
            else:
                if fields[1][:4]=='link':
                    gateway = fields[1][4:].strip('#')
                    gateway_type = 'link'
                else:
                    try:
                        gateway = EthernetMACAddress(fields[1])
                        gateway_type = 'host'
                    except ValueError, emsg:
                        raise RoutingTableError('Error parsing gateway %s: %s' % (fields[1], emsg))

            if isinstance(address, IPv4Address) or isinstance(gateway, IPv4Address):
                if isinstance(address, basestring) and address=='default':
                    address = IPv4Address('default')
                self.ipv4.append(IPv4Route(address, gateway, gateway_type, *fields[2:]))

            if isinstance(address, IPv6Address) or isinstance(gateway, IPv6Address):
                if isinstance(address, basestring) and address=='default':
                    address = IPv6Address('default')
                self.ipv6.append(IPv6Route(address, gateway, gateway_type, *fields[2:]))

    def __parse_address__(self, address):
        try:
            return IPv4Address(address, oldformat=True)

        except ValueError:

            try:
                (address, interface) = address.split('%')
                (interface, mask) = interface.split('/')
            except ValueError:
                interface = None
                mask = None
                pass

            try:
                if mask is not None:
                    return IPv6Address('/'.join([address, mask]))
                else:
                    return IPv6Address(address)
            except IndexError:
                pass
            except ValueError:
                pass

        return None

    def filter(self, **kwargs):
        matches = []
        for route in self.ipv4 + self.ipv6:
            nomatch = False
            for k, v in kwargs.items():

                if k=='gateway' and v=='default':

                    if getattr(route, 'gateway_type')!='route':
                        nomatch = True
                        continue

                    k = 'address'
                    if isinstance(route.address, IPv4Address):
                        v = IPv4Address('default')
                    if isinstance(route.address, IPv6Address):
                        v = IPv6Address('default')

                if not hasattr(route, k) or getattr(route, k)!=v:
                    nomatch = True

            if not nomatch:
                matches.append(route)
        return matches

    def match(self, address):
        matches = []
        for route in self.ipv4 + self.ipv6:
            if route.match(address):
                matches.append(route)
        matches.sort(lambda x, y: cmp(y.address.bitmask, x.address.bitmask))
        return matches

    def best_route(self, address):
        matches = self.match(address)
        if not matches:
            return None
        return matches[0]

class Route(object):
    def __init__(self, address, gateway, gateway_type, interface=None,
                       expire=None, flags=None, refs=None, used=None):
        self.address = address
        self.gateway = gateway
        self.gateway_type = gateway_type
        self.interface = interface
        self.expire = expire
        self.flags = flags is not None and flags or []
        self.refs = refs is not None and refs or 0
        self.used = refs is not None and refs or 0

    def match(self, address):
        if isinstance(address, basestring):
            try:
                address = type(self.address)(address)
            except ValueError:
                return False
        return self.address.addressInNetwork(address)

    @property
    def gateway_mac_address(self):
        if self.gateway_type not in ['route', 'host']:
            return None
        try:
            output = check_output(['arp', '-an'])
        except CalledProcessError, emsg:
            raise RoutingTableError('Error checking arp -rn output: %s' % emsg)
        for l in output.split('\n'):
            try:
                name, addr, at, mac, rest = l.split(None, 4)
                addr = addr.strip('()')
            except ValueError, emsg:
                continue
            if addr==self.gateway:
                return EthernetMACAddress(mac)
        return None

class IPv4Route(Route):
    def __init__(self, address, gateway, gateway_type, flags=None, refs=None,
                       used=None, interface=None, expire=None):
        if flags is not None:
            flags = list(flags)
        if refs is not None:
            refs = int(refs)
        if used is not None:
            used = int(used)
        if expire is not None:
            expire = int(expire)
        Route.__init__(self, address, gateway, gateway_type, refs=refs,
                             used=used, interface=interface, expire=expire)

    def __repr__(self):
        if self.address is None:
            return 'default via %s (%s)' % (self.gateway, self.interface)
        else:
            return '%s via %s (%s)' % (self.address, self.gateway, self.interface)

class IPv6Route(Route):
    def __init__(self, address, gateway, gateway_type,
                       flags=None, interface=None, expire=None):
        if flags is not None:
            flags = list(flags)
        if expire is not None:
            expire = int(expire)
        Route.__init__(self, address, gateway, gateway_type,
                             flags=flags, interface=interface, expire=expire)

    def __repr__(self):
        if self.address is None:
            return 'default via %s (%s)' % (self.gateway, self.interface)
        else:
            return '%s via %s (%s)' % (self.address, self.gateway, self.interface)



#!/usr/bin/env python
"""
HP printer statistics and control module.

Setting of variables not implemented, because my printer does not support
them MIB oids for the ones I was interested about setting.

See example in __main__ how this is used.
"""

import re
import logging

from pysnmp.proto import rfc1902

from seine.snmp import SNMPError, SNMP_VERSIONS
from seine.snmp.client import SNMPClient, SNMPv1Auth, SNMPv2cAuth, SNMPv3Auth

PRINTER_MIB_OID = '.1.3.6.1.2.1.43'
HP_LASERJET_MIB_OID = '.1.3.6.1.4.1.11.2.3.9'

ONLINE_STATUS_VALUES = { 1: 'online', 2: 'offline', 3: 'offline end of job' }

DEVICE_INFO_MAP = {
    'COMMENT':  'comment',
    'DES':      'design',
    'MDL':      'model',
    'MFG':      'manufacturer',
    'MEM':      'memory',
    'CMD':      'languages',
    'CLS':      'class',
    'SN':       'serialnumber',
}

TESTPAGE_MAP = {
    'configuration':    3,
    'errorlog':         7,
    'directorylisting': 8,
    'menumap':          9,
    'usagereport':      10,
    'pcl-fontlist':     350,
    'ps-fontlist':      450,
}
DEFAULT_TESTPAGE = 'configuration'

RESET_TYPE_MAP = {
    'powercycle':       4,
    'reset2nvram':      5,
    'factorydefaults':  6,
}
DEFAULT_RESET_TYPE = 'powercycle'

HP_PRINTER_OID_MAP = {
    'total_pages': {
        'oid': HP_LASERJET_MIB_OID+'.4.2.1.4.1.2.5.0',
        'decode': lambda x: int(x),
    },
    'color_pages': {
        'oid': HP_LASERJET_MIB_OID+'.4.2.1.4.1.2.7.0',
        'decode': lambda x: int(x),
    },
    'duplex_pages': {
        'oid': HP_LASERJET_MIB_OID+'.4.2.1.4.1.2.22.0',
        'decode': lambda x: int(x),
    },
    'config_changes': {
        'oid': PRINTER_MIB_OID + '.5.1.1.1.1',
        'decode': lambda x: int(x),
    },
    'localization': {
        'oid': PRINTER_MIB_OID + '.5.1.1.2.1',
        'decode': lambda x: str(x),
    },
    'reset_status': {
        'oid': PRINTER_MIB_OID + '.5.1.1.3.1',
        'decode': lambda x: int(x),
    },
    'operator_name': {
        'oid': PRINTER_MIB_OID + '.5.1.1.4.1',
        'decode': lambda x: str(x),
    },
    'consolelines': {
        'oid': PRINTER_MIB_OID + '.5.1.1.11.1',
        'decode': lambda x: int(x),
    },
    'consolechars': {
        'oid': PRINTER_MIB_OID + '.5.1.1.12.1',
        'decode': lambda x: int(x),
    },
    'name': {
        'oid': PRINTER_MIB_OID + '.5.1.1.16.1',
        'decode': lambda x: str(x),
    },
    'serialnumber': {
        'oid': PRINTER_MIB_OID + '.5.1.1.16.1',
        'decode': lambda x: str(x),
    },
    'energy_star': {
        'oid': HP_LASERJET_MIB_OID+'.4.2.1.1.1.1.0',
        'decode': lambda x: str(x),
    },
    'sleep_mode': {
        'oid': HP_LASERJET_MIB_OID+'.4.2.1.1.1.2.0',
        'decode': lambda x: int(x),
    },
    'default_copies': {
        'oid': HP_LASERJET_MIB_OID+'.4.2.1.3.3.1.4.0',
        'decode': lambda x: int(x),
    },
    'default_quality': {
        'oid': HP_LASERJET_MIB_OID+'.4.2.1.4.1.6.7.0',
        'decode': lambda x: int(x),
    },
    'res_horizontal': {
        'oid': HP_LASERJET_MIB_OID+'.4.2.1.3.3.1.8.0',
        'decode': lambda x: int(x),
    },
    'res_vertical': {
        'oid': HP_LASERJET_MIB_OID+'.4.2.1.3.3.1.9.0',
        'decode': lambda x: int(x),
    },
    'lines_per_page': {
        'oid': HP_LASERJET_MIB_OID+'.4.2.1.3.3.1.11.0',
        'decode': lambda x: int(x),
    },
    'online_status': {
        'oid': HP_LASERJET_MIB_OID+'.4.2.1.1.2.5.0',
        'decode': lambda x: ONLINE_STATUS_VALUES[int(x)],
    },
    'device_info': {
        'oid': HP_LASERJET_MIB_OID+'.1.1.7.0',
        'decode': lambda x: dict(
            (k in DEVICE_INFO_MAP.keys() and DEVICE_INFO_MAP[k] or k, v.strip()) \
            for k, v in
            [s.split(':') for s in str(x).strip(';').split(';')]
         )
    },
    'supply_names': {
        'oid': PRINTER_MIB_OID + '.11.1.1.6',
        'type': 'tree',
        'decode': lambda k, v: (k,str(v)),
    },
    'supply_max_levels': {
        'oid': PRINTER_MIB_OID + '.11.1.1.8',
        'type': 'tree',
        'decode': lambda k, v: (k,int(v)),
    },
    'supply_levels': {
        'oid': PRINTER_MIB_OID + '.11.1.1.9',
        'type': 'tree',
        'decode': lambda k, v: (k,int(v)),
    },
}

class LaserjetSNMPControl(dict):
    def __init__(self, snmp_client):
        self.client = snmp_client

    def __getattr__(self, attr):
        if attr in HP_PRINTER_OID_MAP.keys():
            d = HP_PRINTER_OID_MAP[attr]
            oid = d['oid']
            dtype = d.get('type', 'get')
            if dtype == 'get':
                value = self.client.get(oid)[1]
                if value == '':
                    return None
                return d['decode'](value)
            elif dtype == 'tree':
                return [d['decode'](k, v) for k, v in self.client.walk(oid).items()]
        raise AttributeError('No such LaserjetSNMPControl attribute: %s' % attr)

    def supply_level_details(self):
        levels = dict(getattr(self, 'supply_names'))
        details = dict((k.split('.')[-1], {'name':levels[k]}) for k, v in levels.items() )
        for oid, maxlevel in getattr(self, 'supply_max_levels'):
            index = oid.split('.')[-1]
            details[index]['max'] = maxlevel
        for oid, level in getattr(self, 'supply_levels'):
            index = oid.split('.')[-1]
            details[index]['level'] = level
        return dict(
            (details[i]['name'], {
                'index': i,
                'max': details[i]['max'],
                'level': details[i]['level'],
                'percent': int(
                    float(details[i]['level'])/float(details[i]['max'])*100
                ),
            }) for i in details.keys()
        )

    def reset(self, resettype=DEFAULT_RESET_TYPE):
        try:
            resettype = rfc1902.Integer(RESET_TYPE_MAP[resettype])
        except KeyError:
            raise ValueError('Invalid reset type: %s' % resettype)
        oid = PRINTER_MIB_OID + '.5.1.1.3.1'
        return self.client.set(oid, resettype)

    def testpage(self, pagetype=DEFAULT_TESTPAGE):
        try:
            pagetype = rfc1902.Integer(TESTPAGE_MAP[pagetype])
        except KeyError:
            raise ValueError('Invalid testpage type: %s' % pagetype)
        oid = HP_LASERJET_MIB_OID + '.4.2.1.1.5.2.0'
        return self.client.set(oid, pagetype)

if __name__ == '__main__':
    import sys

    ljclient = LaserjetSNMPControl(
        SNMPClient(sys.argv[1], SNMPv1Auth(community=sys.argv[2]))
    )
    #ljclient.client.logger.set_level('DEBUG')
    #for k in sorted(HP_PRINTER_OID_MAP.keys()): print k, getattr(ljclient,k)

    for color,details in ljclient.supply_level_details().items():
        print '%12s %s %%' % (color,details['percent'])




"""
Parse netstat -s output fields from linux, OS/X and FreeBSD
"""

import fnmatch
import os
import re
import sys
import time

from datetime import datetime
from subprocess import Popen, PIPE

DATE_FORMAT = '%Y-%m-%d %H:%M:%S'
RE_PROTOCOL = re.compile('^[a-zA-Z0-9]+:$')


COUNTER_REGEXPS = (
    re.compile('^(?P<value>\d+)\s+(?P<name>.*)\s+\([a-z]*\s*(?P<counter>\d+)\s+(?P<countername>[^)]+)\)\s+(?P<name_continued>.*)$'),
    re.compile('^(?P<value>\d+)\s+(?P<name>.*)\s+\([a-z]*\s*(?P<counter>\d+)\s+(?P<countername>[^)]+)\)$'),
    re.compile('^(?P<value>\d+)\s+(?P<name>.*)$'),
    re.compile('^(?P<name>.*):\s+(?P<value>\d+)$'),
)

NETSTAT_PROTOCOL_COUNTERS_LINUX = {

    'icmp': {
        'counter_map': {
            'ICMP messages received':       { 'group': 'received', 'name': 'messages' },
            'input ICMP message failed.':   { 'group': 'received', 'name': 'failed' },
            'ICMP messages sent':           { 'group': 'sent',     'name': 'messages' },
            'ICMP messages failed':         { 'group': 'sent',     'name': 'failed' },
        },

        'histogram_map': {

            'ICMP output histogram:': {
                'group': 'summary',
                'name': 'output',
                'counter_map': {
                    'destination unreachable': { 'group': 'sent', 'name': 'destination_unreachable' },
                    'echo requests': { 'group': 'sent', 'name': 'echo_requests' },
                    'echo replies': { 'group': 'sent', 'name': 'echo_replies' },
                },
            },

            'ICMP input histogram:': {
                'group': 'summary',
                'name': 'input',
                'counter_map': {
                    'destination unreachable': { 'group': 'received', 'name': 'destination_unreachable' },
                    'echo requests': { 'group': 'received', 'name': 'echo_requests' },
                    'echo replies': { 'group': 'received', 'name': 'echo_replies' },
                },
            },

        },

    },

    'icmpmsg': {
        'counter_map': {
            'InType0':  { 'group': 'received', 'name': 'echo_reply' },
            'InType3':  { 'group': 'received', 'name': 'destination_unreachable' },
            'InType8':  { 'group': 'received', 'name': 'echo_request' },
            'OutType0': { 'group': 'sent', 'name': 'echo_reply' },
            'OutType3': { 'group': 'sent', 'name': 'destination_unreachable' },
            'OutType8': { 'group': 'sent', 'name': 'echo_request' },
        },
    },

    'ip': {
        'counter_map': {
            'total packets received': { 'group': 'stats', 'name': 'packets_received' },
            'with invalid addresses': { 'group': 'stats', 'name': 'invalid_address' },
            'forwarded': { 'group': 'stats', 'name': 'forwarded' },
            'requests sent out': { 'group': 'stats', 'name': 'requests_sent' },
            'incoming packets discarded': { 'group': 'stats', 'name': 'incoming_packets_discarded' },
            'incoming packets delivered': { 'group': 'stats', 'name': 'incoming_packets_delivered' },
        },
    },

    'ipext': {
        'counter_map': {
            'InOctets':            { 'group': 'received',  'name': 'octets' },
            'InMcastPkts':         { 'group': 'received',  'name': 'multicast_packets' },
            'InMcastOctets':       { 'group': 'received',  'name': 'multicast_octets' },
            'InNoECTPkts':         { 'group': 'received',  'name': 'no_ect_packets' },
            'OutOctets':           { 'group': 'sent',      'name': 'octets' },
            'OutMcastPkts':        { 'group': 'sent',      'name': 'multicast_packets' },
            'OutMcastOctets':      { 'group': 'sent',      'name': 'multicast_octets' },
        },
    },

    'tcp': {
        'counter_map': {
            'active connections openings':      { 'group': 'connection', 'name': 'active_connection_openings' },
            'passive connection openings':      { 'group': 'connection', 'name': 'passive_connection_openings' },
            'failed connection attempts':       { 'group': 'connection', 'name': 'failed_connection_attemps' },
            'connection resets received':       { 'group': 'connection', 'name': 'connection_resets_received' },
            'connections established':          { 'group': 'connection', 'name': 'connections_established' },
            'resets sent':                      { 'group': 'connection', 'name': 'resets_sent' },
            'segments received':                { 'group': 'segment',    'name': 'received' },
            'segments send out':                { 'group': 'segment', 'name': 'sent_out' },
            'segments retransmited':            { 'group': 'segment', 'name': 'retransmitted' },
            'bad segments received.':           { 'group': 'segment', 'name': 'bad_segments_received' },
        },
    },

    'tcpext': {
        'counter_map': {
            'Invalid SYN cookies received':                                         { 'group': 'errors', 'name': 'invalid_syn_cookies_received' },
            'resets received for embryonic SYN_RECV sockets':                       { 'group': 'errors', 'name': 'resets_for_embryonic_sync_recv_sockets' },
            'packets pruned from receive queue because of socket buffer overrun':   { 'group': 'errors', 'name': 'socket_buffer_overrun_pruned_from_receive_queue' },
            'packets pruned from receive queue':                                    { 'group': 'errors', 'name': 'pruned_from_receive_queue' },
            'ICMP packets dropped because they were out-of-window':                 { 'group': 'stats',  'name': 'dropped_icmp_out_of_window' },
            'TCP sockets finished time wait in fast timer':                         { 'group': 'stats',  'name': 'finish_time_wait_in_fast_timer' },
            'delayed acks sent':                                                    { 'group': 'stats',  'name': 'delayed_acks_sent' },
            'delayed acks further delayed because of locked socket':                { 'group': 'stats',  'name': 'locked_socket_delayed_acks_further_delayed' },
            'Quick ack mode was activated times':                                   { 'group': 'stats',  'name': 'quick_ack_mode_activated' },
            'packets directly queued to recvmsg prequeue.':                         { 'group': 'stats',  'name': 'packets_directly_queued_to_recvmsg_prequeue' },
            'bytes directly received in process context from prequeue':             { 'group': 'stats',  'name': 'bytes_directly_received_in_process_context_from_prequeue' },
            'packet headers predicted':                                             { 'group': 'stats',  'name': 'packet_headers_predicted' },
            'acknowledgments not containing data payload received':                 { 'group': 'stats',  'name': 'received_acknowledgements_not_containing_data' },
            'acknowledgments not containing data received':                         { 'group': 'stats',  'name': 'ack_no_data_received' },
            'predicted acknowledgments':                                            { 'group': 'stats',  'name': 'predicted_acknowledgements' },
            'times recovered from packet loss due to SACK data':                    { 'group': 'stats',  'name': 'recover_from_packet_loss_due_to_sack_data' },
            'congestion windows fully recovered':                                   { 'group': 'stats',  'name': 'cognestion_windows_fully_recovered' },
            'congestion windows partially recovered using Hoe heuristic':           { 'group': 'stats',  'name': 'cognestion_windows_partially_recovered_using_hoe_heuristic' },
            'congestion windows recovered without slow start after partial ack':    { 'group': 'stats',  'name': 'cognestion_windows_recovered_without_slow_start_after_partial_ack' },
            'congestion windows recovered after partial ack':                       { 'group': 'stats',  'name': 'cognestion_windows_recovered_after_partial_ack' },
            'TCP data loss events':                                                 { 'group': 'stats',  'name': 'data_loss_events' },
            'TCPDSACKUndo':                                                         { 'group': 'stats',  'name': 'sack_undo' },
            'TCPLossProbes':                                                        { 'group': 'stats',  'name': 'loss_probes' },
            'TCPLostRetransmit':                                                    { 'group': 'stats',  'name': 'lost_retransmit' },
            'timeouts after SACK recovery':                                         { 'group': 'stats',  'name': 'timeouts_after_sack_recovery' },
            'timeouts in loss state':                                               { 'group': 'stats',  'name': 'timeouts_in_loss_state', },
            'fast retransmits':                                                     { 'group': 'stats',  'name': 'fast_retransmits' },
            'forward retransmits':                                                  { 'group': 'stats',  'name': 'forward_retransmits' },
            'retransmits in slow start':                                            { 'group': 'stats',  'name': 'retransmits_in_slow_start' },
            'other TCP timeouts':                                                   { 'group': 'stats',  'name': 'other_timeouts' },
            'sack retransmits failed':                                              { 'group': 'stats',  'name': 'sack_retransmit_failed' },
            'packets collapsed in receive queue due to low socket buffer':          { 'group': 'stats',  'name': 'packets_collapsed_in_receive_queue_low_socket_buffer' },
            'TCPLossProbeRecovery':                                                 { 'group': 'stats',  'name': 'loss_probe_recovery' },
            'DSACKs sent for old packets':                                          { 'group': 'stats',  'name': 'dsacks_sent_for_old_packet' },
            'DSACKs sent for out of order packets':                                 { 'group': 'stats',  'name': 'dscaks_sent_for_out_of_order_packet' },
            'DSACKs received':                                                      { 'group': 'stats',  'name': 'dsacks_received' },
            'connections reset due to unexpected data':                             { 'group': 'stats',  'name': 'connection_reset_unexpected_data' },
            'connections reset due to early user close':                            { 'group': 'stats',  'name': 'connection_reset_early_user_close' },
            'connections aborted due to timeout':                                   { 'group': 'stats',  'name': 'connections_aborted_due_to_timeout' },
            'TCPDSACKIgnoredOld':                                                   { 'group': 'stats',  'name': 'dsack_ignored_old' },
            'TCPDSACKIgnoredNoUndo':                                                { 'group': 'stats',  'name': 'dsack_ignored_no_undo' },
            'TCPSackShifted':                                                       { 'group': 'stats',  'name': 'sack_shifted' },
            'TCPSackMerged':                                                        { 'group': 'stats',  'name': 'sack_merged' },
            'TCPBacklogDrop':                                                       { 'group': 'stats',  'name': 'sack_backlog_drop' },
            'TCPSpuriousRTOs':                                                      { 'group': 'stats',  'name': 'spurious_rtos' },
            'TCPSackShiftFallback':                                                 { 'group': 'stats',  'name': 'sack_shift_fallback' },
            'TCPRcvCoalesce':                                                       { 'group': 'stats',  'name': 'receive_coalesce' },
            'TCPOFOQueue':                                                          { 'group': 'stats',  'name': 'ofo_queue' },
            'TCPChallengeACK':                                                      { 'group': 'stats',  'name': 'challenge_ack' },
            'TCPSYNChallenge':                                                      { 'group': 'stats',  'name': 'syn_challenge' },
            'TCPAutoCorking':                                                       { 'group': 'stats',  'name': 'auto_corking' },
            'TCPOrigDataSent':                                                      { 'group': 'stats',  'name': 'original_data_sent' },

        },
    },

    'udp': {
        'counter_map': {
            'packets received':                     { 'group': 'stats', 'name': 'packets_received' },
            'packets sent':                         { 'group': 'stats', 'name': 'packets_sent' },
            'packets to unknown port received.':    { 'group': 'errors', 'name': 'packets_received_to_unknown port' },
            'packet receive errors':                { 'group': 'errors', 'name': 'packet_receive_errors' },
            'RcvbufErrors:':                        { 'group': 'errors', 'name': 'receive_buffer_errors' },
        },
    },

}

NETSTAT_PROTOCOL_COUNTERS_BSD = {

    'arp': {

        'counter_map': {
            'ARP requests sent':                                { 'group': 'sent', 'name': 'requests' },
            'ARP replies sent':                                 { 'group': 'sent', 'name': 'replies' },
            'ARP announcements sent':                           { 'group': 'sent', 'name': 'announcements' },
            'ARP conflict probes sent':                         { 'group': 'sent', 'name': 'conflict_probes'},
            'ARP requests received':                            { 'group': 'received', 'name': 'requests' },
            'ARP packets received':                             { 'group': 'received', 'name': 'packets' },
            'ARP replies received':                             { 'group': 'received', 'name': 'replies' },
            'invalid ARP resolve requests':                     { 'group': 'received', 'name': 'invalid_resolve_requests' },
            'total ARP packets received':                       { 'group': 'received', 'name': 'total' },
            'total packets dropped due to lack of memory':      { 'group': 'dropped', 'name': 'lack_of_memory' },
            'total packets dropped due to no ARP entry':        { 'group': 'dropped', 'name': 'no_arp_entry' },
            'total packets dropped during ARP entry removal':   { 'group': 'dropped', 'name': 'during_removal' },
            'ARP entrys timed out':                             { 'group': 'errors', 'name': 'timed_out' },
            'ARP entries timed out':                            { 'group': 'errors', 'name': 'timed_out' },
            'Duplicate IPs seen':                               { 'group': 'errors', 'name': 'duplicate_ip' },
        },

    },

    'icmp': {

        'counter_map': {
            'calls to icmp_error':                              { 'group': 'errors', 'name': 'calls_to_icmp_error' },
            'messages with bad code fields':                    { 'group': 'errors', 'name': 'bad_code_fields' },
            'bad checksums':                                    { 'group': 'errors', 'name': 'bad_checksum' },
            'message with bad length':                          { 'group': 'errors', 'name': 'bad_length' },
            'messages < minimum length':                        { 'group': 'errors', 'name': 'too_short_message' },
            'destination unreachable':                          { 'group': 'stats',  'name': 'destination_unreachable' },
            'echo':                                             { 'group': 'stats',  'name': 'echo' },
            'message responses generated':                      { 'group': 'stats',  'name': 'responses_generated' },
            'multicast echo requests ignored':                  { 'group': 'stats',  'name': 'multicast_echo_request_ignored' },
            'multicast timestamp requests ignored':             { 'group': 'stats',  'name': 'multicast_timestamp_request_ignored' },
            "errors not generated 'cuz old message was icmp":   { 'group': 'stats',  'name': 'error_not_generated_old_icmp' },
        },

        'histogram_map': {

            'Output histogram:': {
                'group': 'summary',
                'name': 'output',
                'counter_map': {
                    'echo reply':               { 'group': 'sent', 'name': 'echo_reply' },
                    'destination unreachable':  { 'group': 'sent', 'name': 'destination_unreachable' },
                }
            },

            'Input histogram:': {
                'group': 'summary',
                'name': 'input',
                'counter_map': {
                    'echo':                     { 'group': 'sent', 'name': 'echo' },
                    'destination unreachable':  { 'group': 'sent', 'name': 'destination_unreachable' },
                }
            },

        },
    },

    'icmp6': {

        'counter_map': {
            'calls to icmp_error': { 'group': 'errors', 'name': 'calls_to_icmp_error' },
        },

        'histogram_map': {

            'Output histogram:': {
                'group': 'summary',
                'name': 'output',
                'counter_map': {
                }
            },

            'Input histogram:': {
                'group': 'summary',
                'name': 'input',
                'counter_map': {
                }
            },

            'Histogram of error messages to be generated:': {
                'group': 'summary',
                'name': 'errors',
                'counter_map': {
                }
            },

        },

    },

    'igmp': {

        'counter_map': {
            'messages received':                                    { 'group': 'received',          'name': 'messages' },
            'membership reports received':                          { 'group': 'received',          'name': 'membership_reports' },
            'membership reports sent':                              { 'group': 'sent',              'name': 'membership_reports' },
            'messages received with too few bytes':                 { 'group': 'errors',            'name': 'messages_too_few_bytes' },
            'messages received with wrong TTL':                     { 'group': 'errors',            'name': 'messages_wrong_ttl' },
            'messages received with bad checksum':                  { 'group': 'errors',            'name': 'bad_checksum' },
            'membership reports received with invalid field(s)':    { 'group': 'errors',            'name': 'membership_reports_with_invalid_fields' },
            'V1/V2 membership queries received':                    { 'group': 'queries_received',  'name': 'v1_v2_membership' },
            'V3 membership queries received':                       { 'group': 'queries_received',  'name': 'v3_membership' },
            'general queries received':                             { 'group': 'queries_received',  'name': 'general' },
            'group queries received':                               { 'group': 'queries_received',  'name': 'group' },
            'group-source queries received':                        { 'group': 'queries_received',  'name': 'group_source' },
            'group-source queries dropped':                         { 'group': 'queries_dropped',   'name': 'group_source' },
        },

    },

    'ip': {

        'group_map': {

            'total packets received': {
                'name': 'packets_received',
                'counter_map': {
                    'bad header checksums':                         { 'group': 'headers',   'name': 'bad_checksums' },
                    'headers checksummed in software':              { 'group': 'headers',   'name': 'checksummed_in_software' },
                    'with size smaller than minimum':               { 'group': 'errors',    'name': 'too_small_size' },
                    'with data size < data length':                 { 'group': 'errors',    'name': 'data_size_less_than_data_length' },
                    'with data size > data length':                 { 'group': 'errors',    'name': 'data_size_more_than_data_length' },
                    'packets forced to software checksum':          { 'group': 'errors',    'name': 'packets_forced_to_sw_checksum' },
                    'packets for unknown/unsupported protocol':     { 'group': 'errors',    'name': 'packet_protocol_unsupported' },
                    'with ip length > max ip packet size':          { 'group': 'errors',    'name': 'ip_length_too_large' },
                    'with header length < data size':               { 'group': 'errors',    'name': 'header_length_more_than_data_size' },
                    'with data length < header length':             { 'group': 'errors',    'name': 'data_length_less_than_header_length' },
                    'with bad options':                             { 'group': 'errors',    'name': 'bad_options' },
                    'with incorrect version number':                { 'group': 'errors',    'name': 'incorrect_version_number' },
                    'packets received for unknown multicast group': { 'group': 'errors',    'name': 'received_for_unknown_multicast_group' },
                    'fragments received':                           { 'group': 'fragments', 'name': 'received' },
                    'reassembled ok':                               { 'group': 'fragments', 'name': 'reassembled_ok' },
                    'packets forwarded':                            { 'group': 'forward',   'name': 'packets_forwarded' },
                    'packets not forwardable':                      { 'group': 'forward',   'name': 'not_forwardable' },
                    'redirects sent':                               { 'group': 'redirect',  'name': 'redirects_sent' },
                    'dropped (dup or out of space)':                { 'group': 'dropped',   'name': 'dup_or_out_of_space' },
                    'dropped after timeout':                        { 'group': 'dropped',   'name': 'after_timeout' },

                }
            },

            'packets sent from this host': {
                'name': 'packets_sent',
                'counter_map': {
                    'packets sent with fabricated ip header':           { 'group': 'headers',   'name': 'fabricated_ip_header' },
                    'headers checksummed in software':                  { 'group': 'headers',   'name': 'checksummed_in_software' },
                    'output packets dropped due to no bufs, etc.':      { 'group': 'dropped',   'name': 'no_buffers_etc' },
                    'output packets discarded due to no route':         { 'group': 'dropped',   'name': 'no_route' },
                    'packets dropped due to no bufs for control data':  { 'group': 'dropped',   'name': 'no_bufs_for_control_data' },
                    'output datagrams fragmented':                      { 'group': 'fragments', 'name': 'output_packets' },
                    'fragments created':                                { 'group': 'fragments', 'name': 'created' },
                    "datagrams that can't be fragmented":               { 'group': 'fragments', 'name': 'datagrams_cant_be_fragmented' },
                    "tunneling packets that can't find gif":            { 'group': 'errors',    'name': 'tunneling_no_gif_found' },
                    'datagrams with bad address in header':             { 'group': 'errors',    'name': 'bad_address_in_header' },
                }
            },

        },
    },

    'ip6': {

        'histogram_map': {

            'Output histogram:': {
                'group': 'summary',
                'name': 'output',
                'counter_map': {
                }
            },

            'Input histogram:': {
                'group': 'summary',
                'name': 'input',
                'counter_map': {
                }
            },

            'Mbuf statistics:': {
                'group': 'stats',
                'name': 'mbuf',
                'counter_map': {
                }
            }
        },
    },

    'ipsec': {

        'counter_map': {
            'inbound packets processed successfully':               { 'group': 'inbound_packets', 'name': 'successfully_processed', },
            'inbound packets violated process security policy':     { 'group': 'inbound_packets', 'name': 'security_policy_violation', },
            'inbound packets with no SA available':                 { 'group': 'inbound_packets', 'name': 'no_sa', },
            'invalid inbound packets':                              { 'group': 'inbound_packets', 'name': 'invalid', },
            'inbound packets failed due to insufficient memory':    { 'group': 'inbound_packets', 'name': 'no_memory', },
            'inbound packets failed getting SPI':                   { 'group': 'inbound_packets', 'name': 'error_getting_spi', },
            'inbound packets failed on AH replay check':            { 'group': 'inbound_packets', 'name': 'ah_replay_check_fail', },
            'inbound packets failed on ESP replay check':           { 'group': 'inbound_packets', 'name': 'esp_reaply_check_fail', },
            'inbound packets considered authentic':                 { 'group': 'inbound_packets', 'name': 'authentic', },
            'inbound packets failed on authentication':             { 'group': 'inbound_packets', 'name': 'authentication_failure', },
            'outbound packets processed successfully':              { 'group': 'outbound_packets', 'name': 'successfully_processed', },
            'outbound packets violated process security policy':    { 'group': 'outbound_packets', 'name': 'security_policy_violation', },
            'outbound packets with no SA available':                { 'group': 'outbound_packets', 'name': 'no_sa', },
            'invalid outbound packets':                             { 'group': 'outbound_packets', 'name': 'invalid', },
            'outbound packets failed due to insufficient memory':   { 'group': 'outbound_packets', 'name': 'no_memory', },
            'outbound packets with no route':                       { 'group': 'outbound_packets', 'name': 'no_route', },
        },

        'histogram_map': {

            'ESP output histogram:': {
                'group': 'summary',
                'name': 'output',
                'counter_map': {
                }
            },

            'ESP input histogram:': {
                'group': 'summary',
                'name': 'input',
                'counter_map': {
                }
            },
        },
    },

    'ipsec6': {

        'counter_map': {
            'inbound packets processed successfully':               { 'group': 'inbound_packets', 'name': 'successfully_processed', },
            'inbound packets violated process security policy':     { 'group': 'inbound_packets', 'name': 'security_policy_violation', },
            'inbound packets with no SA available':                 { 'group': 'inbound_packets', 'name': 'no_sa', },
            'invalid inbound packets':                              { 'group': 'inbound_packets', 'name': 'invalid', },
            'inbound packets failed due to insufficient memory':    { 'group': 'inbound_packets', 'name': 'no_memory', },
            'inbound packets failed getting SPI':                   { 'group': 'inbound_packets', 'name': 'error_getting_spi', },
            'inbound packets failed on AH replay check':            { 'group': 'inbound_packets', 'name': 'ah_replay_check_fail', },
            'inbound packets failed on ESP replay check':           { 'group': 'inbound_packets', 'name': 'esp_reaply_check_fail', },
            'inbound packets considered authentic':                 { 'group': 'inbound_packets', 'name': 'authentic', },
            'inbound packets failed on authentication':             { 'group': 'inbound_packets', 'name': 'authentication_failure', },
            'outbound packets processed successfully':              { 'group': 'outbound_packets', 'name': 'successfully_processed', },
            'outbound packets violated process security policy':    { 'group': 'outbound_packets', 'name': 'security_policy_violation', },
            'outbound packets with no SA available':                { 'group': 'outbound_packets', 'name': 'no_sa', },
            'invalid outbound packets':                             { 'group': 'outbound_packets', 'name': 'invalid', },
            'outbound packets failed due to insufficient memory':   { 'group': 'outbound_packets', 'name': 'no_memory', },
            'outbound packets with no route':                       { 'group': 'outbound_packets', 'name': 'no_route', },
        },
    },

    'kctl': {

        'counter_map': {
            'total kernel control modules registered':      { 'group': 'kernel', 'name': 'total_control_modules_registered', },
            'current kernel control modules registered':    { 'group': 'kernel', 'name': 'current_control_modules_registered', },
            'current kernel control sockets':               { 'group': 'kernel', 'name': 'current_control_sockets' },
            'kernel control generation count':              { 'group': 'kernel', 'name': 'control_generations' },
            'connection attempts':                          { 'group': 'stats', 'name': 'connection_attempts' },
            'connection failures':                          { 'group': 'stats', 'name': 'connection_failures' },
            'send failures':                                { 'group': 'stats', 'name': 'send_failures' },
            'send list failures':                           { 'group': 'stats', 'name': 'send_list_failures' },
            'enqueus failures':                             { 'group': 'stats', 'name': 'enqueue_failures' },
            'packets dropped due to full socket buffers':   { 'group': 'dropped', 'name': 'full_socket_buffers' },
        },
    },

    'kevt': {

        'counter_map': {
            'current kernel control sockets':               { 'group': 'kernel', 'name': 'current_control_sockets' },
            'kernel control generation count':              { 'group': 'kernel', 'name': 'control_generation_count' },
            'message posteds':                              { 'group': 'kernel', 'name': 'messages_posted' },
            'bad vendor failures':                          { 'group': 'errors', 'name': 'bad_vendor' },
            'message too big failures':                     { 'group': 'errors', 'name': 'message_too_big' },
            'out of memeory failures':                      { 'group': 'errors', 'name': 'out_of_memory' },
            'messages dropped due to full socket buffers':  { 'group': 'errors', 'name': 'full_socket_buffers' },
        },
    },

    'mptcp': {

        'counter_map': {
            'data packets sent':                            { 'group': 'sent', 'name': 'packets' },
            'data bytes sent':                              { 'group': 'sent', 'name': 'bytes' },
            'data packets received':                        { 'group': 'received', 'name': 'packets' },
            'data bytes received':                          { 'group': 'received', 'name': 'bytes' },
            'subflow switches':                             { 'group': 'stats', 'name': 'subflow_switch' },
            'times the MPTCP subflow window was reduced':   { 'group': 'stats', 'name': 'subflow_window_reduce' },
            'packets with an invalid MPCAP option':         { 'group': 'errors', 'name': 'invalid_mpcap_option' },
            'packets with an invalid MPJOIN option':        { 'group': 'errors', 'name': 'invalid_mpjoin_option' },
            'times primary subflow fell back to TCP':       { 'group': 'errors', 'name': 'primary_subflow_tcp_fallback' },
            'times secondary subflow fell back to TCP':     { 'group': 'errors', 'name': 'secondary_subflow_tcp_fallback' },
            'DSS option drops':                             { 'group': 'errors', 'name': 'dss_option_drop' },
            'bad DSS checksums':                            { 'group': 'errors', 'name': 'bad_dss_checksum' },
            'times received out of order data':             { 'group': 'errors', 'name': 'out_of_order_data_received' },
            'other invalid MPTCP options':                  { 'group': 'errors', 'name': 'other_invalid_option_drop' },
        },
    },

    'pfkey': {

        'group_map': {

            'requests sent to userland': {
                'name': 'packets_sent',
                'counter_map': {
                    'bytes sent to userland':                       { 'group': 'sent', 'name': 'bytes_to_userland' },
                    'messages with invalid length field':           { 'group': 'errors', 'name': 'invalid_length' },
                    'messages with invalid version field':          { 'group': 'errors', 'name': 'invalid_version' },
                    'messages with invalid message type field':     { 'group': 'errors', 'name': 'invalid_message_type' },
                    'messages too short':                           { 'group': 'errors', 'name': 'too_short' },
                    'messages with memory allocation failure':      { 'group': 'errors', 'name': 'memory_allocation_fail' },
                    'messages with duplicate extension':            { 'group': 'errors', 'name': 'duplicate_extension' },
                    'messages with invalid extension type':         { 'group': 'errors', 'name': 'invalid_extension_type' },
                    'messages with invalid sa type':                { 'group': 'errors', 'name': 'invalid_sa_type' },
                    'messages with invalid address extension':      { 'group': 'errors', 'name': 'invalid_address_extension' },
                },
                'histogram_map': {
                    'histogram by message type:': {
                        'group': 'histogram',
                        'name': 'histogram_by_message_type',
                        'counter_map': {
                            'getspi':       { 'group': 'send', 'name': 'getspi' },
                            'update':       { 'group': 'send', 'name': 'update' },
                            'add':          { 'group': 'send', 'name': 'add' },
                            'delete':       { 'group': 'send', 'name': 'delete' },
                            'acquire':      { 'group': 'send', 'name': 'acquire' },
                            'register':     { 'group': 'send', 'name': 'register' },
                            'dump':         { 'group': 'send', 'name': 'dump' },
                            'x_spdadd':     { 'group': 'send', 'name': 'x_spdadd' },
                            'x_spddelete':  { 'group': 'send', 'name': 'x_spddelete' },
                            'x_spddump':    { 'group': 'send', 'name': 'x_spddump' },
                        },
                    },
                },
            },

            'requests sent from userland': {
                'name': 'packets_received',
                'counter_map': {
                    'bytes sent from userland':                     { 'group': 'sent', 'name': 'bytes_from_userland' },
                    'messages with memory allocation failure':      { 'group': 'errors', 'name': 'memory_allocationfailure' },
                    'messages toward single socket':                { 'group': 'sockets', 'name': 'to_single_socket' },
                    'messages toward all sockets':                  { 'group': 'sockets', 'name': 'to_all_sockets' },
                    'messages toward registered sockets':           { 'group': 'sockets', 'name': 'to_registered_sockets' },
                },
                'histogram_map': {
                    'histogram by message type:':                    {
                        'group': 'histogram',
                        'name': 'histogram_by_message_type',
                        'counter_map': {
                            'getspi':       { 'group': 'received', 'name': 'getspi' },
                            'update':       { 'group': 'received', 'name': 'update' },
                            'add':          { 'group': 'received', 'name': 'add' },
                            'expire':       { 'group': 'send', 'name': 'expire' },
                            'delete':       { 'group': 'received', 'name': 'delete' },
                            'acquire':      { 'group': 'received', 'name': 'acquire' },
                            'register':     { 'group': 'received', 'name': 'register' },
                            'dump':         { 'group': 'received', 'name': 'dump' },
                            'x_spdadd':     { 'group': 'received', 'name': 'x_spdadd' },
                            'x_spddelete':  { 'group': 'received', 'name': 'x_spddelete' },
                            'x_spddump':    { 'group': 'received', 'name': 'x_spddump' },
                        },
                    },
                },
            },

        },

    },

    'rip6': {
        'counter_map': {},
        'group_map': {},
        'histogram_map': {},
    },

    'sctp': {
        'counter_map': {
            "packet shorter than header":                                   { 'group': 'errors', 'name': 'packet_shorter_than_header' },
            "checksum error":                                               { 'group': 'errors', 'name': 'checksum_error' },
            "no endpoint for port":                                         { 'group': 'errors', 'name': 'no_endpoint_for_port' },
            "bad v-tag":                                                    { 'group': 'errors', 'name': 'bad_vtag' },
            "bad SID":                                                      { 'group': 'errors', 'name': 'bad_sid' },
            "no memory":                                                    { 'group': 'errors', 'name': 'no_memory' },
            "number of multiple FR in a RTT window":                        { 'group': 'stats', 'name': 'multiple_fr_in_rtt_window' },
            "RFC813 allowed sending":                                       { 'group': 'stats', 'name': 'rfc813_allowed_sending' },
            "RFC813 does not allow sending":                                { 'group': 'stats', 'name': 'rfc813_does_not_allow_sending' },
            "times max burst prohibited sending":                           { 'group': 'stats', 'name': 'max_burst_prohibited_sending' },
            "look ahead tells us no memory in interface":                   { 'group': 'stats', 'name': 'lookahead_no_memory_in_interface' },
            "numbers of window probes sent":                                { 'group': 'stats', 'name': 'window_probes_sent' },
            "times an output error to clamp down on next user send":        { 'group': 'stats', 'name': 'output_error_to_clamp_down_next_user_send' },
            "times sctp_senderrors were caused from a user":                { 'group': 'stats', 'name': 'sctp_senderrors_caused_from_user' },
            "number of in data drops due to chunk limit reached":           { 'group': 'stats', 'name': 'data_drops_due_to_chunk_limit' },
            "number of in data drops due to rwnd limit reached":            { 'group': 'stats', 'name': 'data_drops_due_to_rwnd_limit' },
            "times a ECN reduced the cwnd":                                 { 'group': 'stats', 'name': 'ecn_reduced_cwnd' },
            "used express lookup via vtag":                                 { 'group': 'stats', 'name': 'express_lookup_via_vtag' },
            "collision in express lookup":                                  { 'group': 'stats', 'name': 'collision_in_express_lookup' },
            "times the sender ran dry of user data on primary":             { 'group': 'stats', 'name': 'sender_ran_dry_of_user_data_on_primary' },
            "sacks the slow way":                                           { 'group': 'stats', 'name': 'slow_way_sacks' },
            "window update only sacks sent":                                { 'group': 'stats', 'name': 'window_update_only_sacks_sent' },
            "sends with sinfo_flags !=0":                                   { 'group': 'stats', 'name': 'sends_with_nonzero_sinfo_flags' },
            "unordered sends":                                              { 'group': 'stats', 'name': 'unordered_sends' },
            "sends with EOF flag set":                                      { 'group': 'stats', 'name': 'sends_with_eof_flag_set' },
            "sends with ABORT flag set":                                    { 'group': 'stats', 'name': 'sends_with_abort_flag_set' },
            "times protocol drain called":                                  { 'group': 'stats', 'name': 'protocol_drains_called' },
            "times we did a protocol drain":                                { 'group': 'stats', 'name': 'protocol_drains_done' },
            "times recv was called with peek":                              { 'group': 'stats', 'name': 'recv_called_with_peek' },
            "cached chunks used":                                           { 'group': 'stats', 'name': 'cached_chunks_used' },
            "cached stream oq's used":                                      { 'group': 'stats', 'name': 'cache_stream_oq_used' },
            "unread messages abandonded by close":                          { 'group': 'stats', 'name': 'unread_messages_abandoned_by_close' },
            "send burst avoidance, already max burst inflight to net":      { 'group': 'stats', 'name': 'send_burst_avoidance' },
            "send cwnd full avoidance, already max burst inflight to net":  { 'group': 'stats', 'name': 'send_cwnd_full_avoidance' },
            "number of map array over-runs via fwd-tsn's":                  { 'group': 'stats', 'name': 'map_array_overruns_via_fwd_tsn' },
        },
        'group_map': {
            'input packets': {
                'name': 'input_packets',
                'counter_map': {
                    'datagrams':                        { 'group': 'stats', 'name': 'datagrams' },
                    'packets that had data':            { 'group': 'stats', 'name': 'packets_with_data' },
                    'input SACK chunks':                { 'group': 'stats', 'name': 'input_sack_chunks' },
                    'input DATA chunks':                { 'group': 'stats', 'name': 'input_data_chunks' },
                    'duplicate DATA chunks':            { 'group': 'stats', 'name': 'duplicate_data_chunks' },
                    'input HB chunks':                  { 'group': 'stats', 'name': 'input_hb_chunks' },
                    'HB-ACK chunks':                    { 'group': 'stats', 'name': 'hb_ack_chunks' },
                    'input ECNE chunks':                { 'group': 'stats', 'name': 'input_ecne_chunks' },
                    'input AUTH chunks':                { 'group': 'stats', 'name': 'input_auth_chunks' },
                    'chunks missing AUTH':              { 'group': 'stats', 'name': 'chunks_without_auth' },
                    'invalid HMAC ids received':        { 'group': 'stats', 'name': 'invalid_hmac_ids_received' },
                    'invalid secret ids received':      { 'group': 'stats', 'name': 'invalid_secret_ids_received' },
                    'auth failed':                      { 'group': 'stats', 'name': 'auth_failed' },
                    'fast path receives all one chunk': { 'group': 'stats', 'name': 'fast_path_receives_all_one_chunk' },
                    'fast path multi-part data':        { 'group': 'stats', 'name': 'fast_path_multipart_data' },

                },
            },
            'output packets': {
                'name': 'output_packets',
                'counter_map': {
                    "output SACKs":                                     { 'group': 'stats', 'name': 'output_sacks', },
                    "output DATA chunks":                               { 'group': 'stats', 'name': 'output_data_chunks', },
                    "retransmitted DATA chunks":                        { 'group': 'stats', 'name': 'retransmitted_data_chunks', },
                    "fast retransmitted DATA chunks":                   { 'group': 'stats', 'name': 'fast_retransmitted_data_chunks', },
                    "FR's that happened more than once to same chunk":  { 'group': 'stats', 'name': 'fast_retransmits_more_than_once_to_same_chunk', },
                    "output HB chunks":                                 { 'group': 'stats', 'name': 'output_hb_chunks', },
                    "output ECNE chunks":                               { 'group': 'stats', 'name': 'output_ecne_chunks', },
                    "output AUTH chunks":                               { 'group': 'stats', 'name': 'output_auth_chunks', },
                    "ip_output error counter":                          { 'group': 'stats', 'name': 'ip_output_errors', },

                },
            },
        },
        'histogram_map': {
            'Packet drop statistics:': {
                'group': 'histogram',
                'name': 'packet_drop_statistics',
                'counter_map': {
                    "from middle box":                      { 'group': 'stats', 'name': 'from_middle_box' },
                    "from end host":                        { 'group': 'stats', 'name': 'from_end_box' },
                    "with data":                            { 'group': 'stats', 'name': 'with_data' },
                    "non-data, non-endhost":                { 'group': 'stats', 'name': 'non_data_non_endhost' },
                    "non-endhost, bandwidth rep only":      { 'group': 'stats', 'name': 'non_data_only_bandwidth_rep' },
                    "not enough for chunk header":          { 'group': 'stats', 'name': 'not_enough_for_chunk_header' },
                    "not enough data to confirm":           { 'group': 'stats', 'name': 'not_enough_data_to_confirm' },
                    "where process_chunk_drop said break":  { 'group': 'stats', 'name': 'where_process_chunk_stop_said_break' },
                    "failed to find TSN":                   { 'group': 'stats', 'name': 'failed_to_find_tsn' },
                    "attempt reverse TSN lookup":           { 'group': 'stats', 'name': 'attempt_reverse_tsn_lookup' },
                    "e-host confirms zero-rwnd":            { 'group': 'stats', 'name': 'ehost_confirm_zero_rewind' },
                    "midbox confirms no space":             { 'group': 'stats', 'name': 'midbox_confirm_no_space' },
                    "data did not match TSN":               { 'group': 'stats', 'name': 'data_did_not_match_tsn' },
                    "TSN's marked for Fast Retran":         { 'group': 'stats', 'name': 'tsn_makred_for_fast_retran' },
                },
            },
            'Timeouts:': {
                'group': 'histogram',
                'name': 'timeouts',
                'counter_map': {
                    'iterator timers fired':                { 'group': 'stats', 'name': 'iterator_timer_fired' },
                    'T3 data time outs':                    { 'group': 'stats', 'name': 't3_data_timeouts' },
                    'window probe (T3) timers fired':       { 'group': 'stats', 'name': 'window_probe_t3_timers_fired' },
                    'INIT timers fired':                    { 'group': 'stats', 'name': 'init_timer_fired' },
                    'sack timers fired':                    { 'group': 'stats', 'name': 'sack_timer_fired' },
                    'shutdown timers fired':                { 'group': 'stats', 'name': 'shutdown_timer_fired' },
                    'heartbeat timers fired':               { 'group': 'stats', 'name': 'heartbeat_timer_fired' },
                    'a cookie timeout fired':               { 'group': 'stats', 'name': 'cookie_timeout_fired' },
                    'an endpoint changed its cookiesecret': { 'group': 'stats', 'name': 'endpoint_changed_cookiesecret' },
                    'PMTU timers fired':                    { 'group': 'stats', 'name': 'pmtu_timer_fired' },
                    'shutdown ack timers fired':            { 'group': 'stats', 'name': 'shutdown_ack_timer_fired' },
                    'shutdown guard timers fired':          { 'group': 'stats', 'name': 'shutdown_guard_timer_fired' },
                    'stream reset timers fired':            { 'group': 'stats', 'name': 'stream_reset_timer_fired' },
                    'early FR timers fired':                { 'group': 'stats', 'name': 'early_fr_timer_fired' },
                    'an asconf timer fired':                { 'group': 'stats', 'name': 'asconf_timer_fired' },
                    'auto close timer fired':               { 'group': 'stats', 'name': 'auto_close_timer_fired' },
                    'asoc free timers expired':             { 'group': 'stats', 'name': 'asoc_free_timer_expired' },
                    'inp free timers expired':              { 'group': 'stats', 'name': 'inp_free_timer_expired' },
                },
            },
        },
    },

    'tcp': {

        'counter_map': {
            'connection requests':                              { 'group': 'received', 'name': 'connection_request', },
            'connection accepts':                               { 'group': 'received', 'name': 'connection_accept', },
            'SACK options (SACK blocks) received':              { 'group': 'received', 'name': 'sack_options' },
            'bad connection attempts':                          { 'group': 'errors', 'name': 'connection_attempt_bad', },
            'listen queue overflows':                           { 'group': 'errors', 'name': 'listen_queue_overflow', },
            'connections established (including accepts)':      { 'group': 'stats', 'name': 'established' },
            'segments updated rtt (of 1172312 attempts)':       { 'group': 'stats', 'name': 'segments_updated_rtt' },
            'correct ACK header predictions':                   { 'group': 'stats', 'name': 'correct_ack_header_predictions' },
            'correct data packet header predictions':           { 'group': 'stats', 'name': 'correct_data_packet_header_predictions' },
            'SACK recovery episodes':                           { 'group': 'stats', 'name': 'sack_recovery_episodes' },
            'SACK scoreboard overflow':                         { 'group': 'stats', 'name': 'sack_scoreboard_overflow' },
            'times cumulative ack advanced along with SACK':    { 'group': 'stats', 'name': 'cumulative_ack_advanced_with_sack' },
            'segment rexmits in SACK recovery episodes':        { 'group': 'stats', 'name': 'segment_rexmit_in_sack_recovery_episodes' },
            'byte rexmits in SACK recovery episodes':           { 'group': 'stats', 'name': 'byte_rexmit_in_sack_recovery_episodes' },
            'limited transmits done':                           { 'group': 'stats', 'name': 'limited_transmits' },
            'early retransmits done':                           { 'group': 'stats', 'name': 'early_transmits' },
            'SACK options (SACK blocks) sent':                  { 'group': 'sent',  'name': 'sack_options' },
            'embryonic connections dropped':                    { 'group': 'dropped', 'name': 'embryonic_connections' },
        },

        'group_map': {

            'packets sent': {
                'name': 'packets_sent',
                'counter_map': {
                    'data packets':                         { 'group': 'sent', 'name': 'packets' },
                    'data packets retransmitted':           { 'group': 'sent', 'name': 'packets_retransmitted' },
                    'resends initiated by MTU discovery':   { 'group': 'sent', 'name': 'resend_by_mtu_discovery' },
                    'ack-only packets':                     { 'group': 'sent', 'name': 'ack_only' },
                    'URG only packets':                     { 'group': 'sent', 'name': 'urg_only' },
                    'window probe packets':                 { 'group': 'sent', 'name': 'window_probe' },
                    'window update packets':                { 'group': 'sent', 'name': 'window_update' },
                    'control packets':                      { 'group': 'sent', 'name': 'control_packets' },
                    'data packets sent after flow control': { 'group': 'sent', 'name': 'data_packets_after_flow_control' },
                    'checksummed in software': {
                        'group': 'stats',
                        'name': 'checksummed_in_software',
                        'counter_map': {
                            'segments over IPv4': { 'group': 'segments', 'name': 'ipv4' },
                            'segments over IPv6': { 'group': 'segments', 'name': 'ipv6' },
                        }
                    },
                },
                'subgroups': (
                    'checksummed_in_software',
                ),
            },

            'packets received': {
                'name': 'packets_received',
                'counter_map': {
                    'acks':                                         { 'group': 'received',  'name': 'acks' },
                    'duplicate acks':                               { 'group': 'received',  'name': 'duplicate_acks' },
                    'acks for unsent data':                         { 'group': 'received',  'name': 'acks_for_unsent' },
                    'packets received in-sequence':                 { 'group': 'received',  'name': 'packets_in_sequence' },
                    'completely duplicate packets':                 { 'group': 'duplicate', 'name': 'completely_duplicate' },
                    'packets with some dup. data':                  { 'group': 'duplicate', 'name': 'some_duplicate_data' },
                    'old duplicate packets':                        { 'group': 'duplicate', 'name': 'old' },
                    'received packets dropped due to low memory':   { 'group': 'errors',    'name': 'low_memory' },
                    'out-of-order packets':                         { 'group': 'errors',    'name': 'out_of_order' },
                    'packets of data after window':                 { 'group': 'errors',    'name': 'data_after_window' },
                    'packets received after close':                 { 'group': 'errors',    'name': 'received_after_close' },
                    'bad resets':                                   { 'group': 'errors',    'name': 'bad_reset' },
                    'discarded for bad checksums':                  { 'group': 'discarded', 'name': 'bad_checksum_discarded' },
                    'discarded for bad header offset fields':       { 'group': 'discarded', 'name': 'bad_header_offset_fields' },
                    'discarded because packet too short':           { 'group': 'discarded', 'name': 'packet_too_short' },
                    'window probes':                                { 'group': 'stats',     'name': 'window_probes' },
                    'window update packets':                        { 'group': 'stats',     'name': 'window_update' },
                    'checksummed in software': {
                        'group': 'stats',
                        'name': 'checksummed_in_software',
                        'counter_map': {
                            'segments over IPv4': { 'group': 'segments', 'name': 'ipv4' },
                            'segments over IPv6': { 'group': 'segments', 'name': 'ipv6' },
                        }
                    },
                },
                'subgroups': (
                    'checksummed_in_software',
                ),
            },

            'retransmit timeouts': {
                'name': 'retransmit_timeouts',
                'counter_map': {
                    'connections dropped by rexmit timeout':        { 'group': 'dropped', 'name': 'after_retransmit_timeout' },
                    'connections dropped after retransmitting FIN': { 'group': 'dropped', 'name': 'after_fin_retransmit' },
                }
            },

            'keepalive timeouts': {
                'name':  'keepalive_timeouts',
                'counter_map': {
                    'keepalive probes sent':                    { 'group': 'sent',    'name': 'probes_sent' },
                    'connections dropped by keepalive':         { 'group': 'dropped', 'name': 'connections_dropped_by_keepalive' },
                }
            },

            'persist timeouts': {
                'name': 'persist_timeouts',
                'counter_map': {
                    'connections dropped by persist timeout':   { 'group': 'dropped', 'name': 'persist_timeout' },
                }
            },

            'LRO coalesced packets': {
                'name': 'lro_coalesced_packets',
                'counter_map': {
                    'collisions in LRO flow table':             { 'group': 'stats', 'name': 'lro_flow_table_collisions' },
                    'times LRO flow table was full':            { 'group': 'stats', 'name': 'lro_flow_table_full_count' },
                    'times LRO coalesced 2 packets':            { 'group': 'stats', 'name': 'lro_coalesce_2_count' },
                    'times LRO coalesced 3 or 4 packets':       { 'group': 'stats', 'name': 'lro_coalesce_3_4_count' },
                    'times LRO coalesced 5 or more packets':    { 'group': 'stats', 'name': 'lro_coalesce_5_count' },
                }
            },

            'connections negotiated ECN': {
                'name': 'ecn_negotiated_connections',
                'counter_map': {
                    'times congestion notification was sent using ECE': { 'group': 'stats', 'name': 'ece_cognestion_notifications' },
                    'times CWR was sent in response to ECE':            { 'group': 'stats', 'name': 'ece_response_with_cwr' },
                }
            },

            'times packet reordering was detected on a connection': {
                'name': 'packet_reordering_detected',
                'counter_map': {
                    'times transmitted packets were reordered':                 { 'group': 'stats', 'name': 'transmitted_packets_reordered' },
                    'times fast recovery was delayed to handle reordering':     { 'group': 'stats', 'name': 'fast_recovery_delayed' },
                    'times retransmission was avoided by delaying recovery':    { 'group': 'stats', 'name': 'delaying_recovery_avoided_retransmission' },
                    'retransmissions not needed':                               { 'group': 'stats', 'name': 'not_needed' },
                }
            },

        },

    },

    'udp': {
        'counter_map': {},
        'group_map': {

            'datagrams received': {
                'name': 'datagrams_received',
                'counter_map': {
                    'delivered':                                    { 'group': 'stats', 'name': 'delivered' },
                    'times multicast source filter matched':        { 'group': 'stats', 'name': 'multicast_filter_match' },
                    'with incomplete header':                       { 'group': 'stats', 'name': 'incomplete_header' },
                    'with bad data length field':                   { 'group': 'stats', 'name': 'bad_data_length_field' },
                    'with bad checksum':                            { 'group': 'stats', 'name': 'bad_checksum' },
                    'with no checksum':                             { 'group': 'stats', 'name': 'no_checksum' },
                    'not for hashed pcb':                           { 'group': 'stats', 'name': 'not_hashed_pcb' },
                    'broadcast/multicast datagrams undelivered':    { 'group': 'stats', 'name': 'undelivered_broadcast_multicast' },
                    'dropped due to no socket':                     { 'group': 'dropped', 'name': 'no_socket' },
                    'dropped due to full socket buffers':           { 'group': 'dropped', 'name': 'socket_buffers_full' },
                    'checksummed in software': {
                        'group': 'stats',
                        'name': 'checksummed_in_software',
                        'counter_map': {
                            'datagrams over IPv4': { 'group': 'checksummed_datagrams', 'name': 'ipv4' },
                            'datagrams over IPv6': { 'group': 'checksummed_datagrams', 'name': 'ipv6' },
                        }
                    },
                },
                'subgroups': (
                    'checksummed_in_software',
                ),
            },

            'datagrams output': {
                'name': 'datagrams_sent',
                'counter_map': {
                    'checksummed in software': {
                        'group': 'stats',
                        'name': 'checksummed_in_software',
                        'counter_map': {
                            'datagrams over IPv4': { 'group': 'checksummed_datagrams', 'name': 'ipv4' },
                            'datagrams over IPv6': { 'group': 'checksummed_datagrams', 'name': 'ipv6' },
                        }
                    },
                },
                'subgroups': (
                    'checksummed_in_software',
                ),
            },

        },
        'histogram_map': {},
    },

}


class NetstatStatisticsError(Exception):
    pass


class Counter(object):
    """Netstat counter

    Instance of output couunter
    """
    def __init__(self, protocol, parent, group, name, value=None, counter_map={}):
        self.protocol = protocol
        self.parent = parent
        self.group = group
        self.name = name.rstrip(':')
        self.details = {}

        if value is not None:
            self.value = int(value)
        else:
            self.value = value

        self.counter_map = counter_map
        self.counters = []

    @property
    def path(self):
        """Path to counter

        Dot separated path to counter
        """
        return '.'.join([self.parent.path, self.group, self.name])

    def __repr__(self):
        return self.path

    def __cmp__(self, other):
        for key in ('protocol', 'group', 'name', ):
            a = getattr(self, key)
            b = getattr(other, key)
            if a != b:
                return cmp(a, b)
        return 0

    def add_counter(self, name, value):
        try:
            config = self.counter_map[name]
            counter = Counter(self.protocol, self, config['group'], config['name'], value, counter_map=config.get('counter_map', {}))
            self.counters.append(counter)
            return counter

        except KeyError:
            return None

    def as_dict(self):
        data = {
            'name': self.name,
            'path': self.path,
            'value': self.value,
        }

        if self.details:
            data['details'] = self.details

        if self.counters:
            data['counters'] = [counter.as_dict() for counter in self.counters]

        return data


class Group(Counter):
    """Group of counters

    More detailed counters grouped under a higher level counter value.
    """
    def __init__(self, protocol, name, value, counter_map, subgroups, histogram_map):
        Counter.__init__(self, protocol, protocol, name, name, value, counter_map)

        self.subgroups = subgroups

        self.histogram_map = histogram_map
        self.histograms = []

        self.current_histogram = None
        self.current_subgroup = None

    @property
    def path(self):
        return '.'.join([self.protocol.path, self.name])

    def __repr__(self):
        return '%s.%s group %s value %s' % (self.protocol.name, self.group, self.name, self.value)

    def set_histogram(self, name):
        try:
            details = self.histogram_map[name]
        except KeyError:
            raise NetstatStatisticsError('Invalid %s histogram: %s' % (self.name, name))

        histogram = Histogram(self, self, details['group'], details['name'], counter_map=details['counter_map'])
        self.histograms.append(histogram)
        self.current_histogram = histogram
        return histogram

    def add_counter(self, name, value):
        if self.current_subgroup:
            counter = self.current_subgroup.add_counter(name, value)
            if counter is not None:
                return counter
            else:
                self.current_subgroup = None
                self.current_histogram = None

        counter = Counter.add_counter(self, name, value)

        if counter is not None:
            if counter.name in self.subgroups:
                self.current_subgroup = counter
                self.current_histogram = None

        elif self.current_histogram is not None:
            self.current_histogram.add_counter(name, value)

        return counter

    def as_dict(self):
        data = Counter.as_dict(self)
        if self.histograms:
            data['histograms'] = [histogram.as_dict() for histogram in self.histograms]
        return data


class Histogram(Counter):
    """Histogram group

    Group of counter where the group has no counter, just title.
    """
    def __init__(self, protocol, parent, group, name, counter_map):
        Counter.__init__(self, protocol, parent, group, name, counter_map=counter_map)

    def __repr__(self):
        return '%s.%s %s %s' % (self.protocol.name, self.group, self.name)

    def add_counter(self, name, value):
        counter = Counter.add_counter(self, name, value)

        if counter is not None:
            return counter


class Protocol(dict):
    """Protocol

    Protocol reported by netstat -s, with lowercase name
    """
    def __init__(self, name, counter_map=None, group_map=None, histogram_map=None):
        self.name = name
        self.counter_map = counter_map
        self.group_map = group_map
        self.histogram_map = histogram_map

        self.groups = []
        self.histograms = []

        self.current_group = None
        self.current_histogram = None

    def __repr__(self):
        return self.name

    def __cmp__(self, other):
        return cmp(self.name, other.name)

    @property
    def path(self):
        return self.name

    @property
    def all_counters(self):
        counters = []
        counters.extend(counter for group in self.values() for counter in group)

        for histogram in self.histograms:
            counters.extend(counter for counter in histogram.counters)

        for group in self.groups:
            for counter in group.counters:
                counters.append(counter)
                counters.extend(counter for counter in counter.counters)

            for histogram in group.histograms:
                counters.extend(counter for counter in histogram.counters)

        return counters

    def keys(self):
        return [key for key in sorted(dict.keys(self))]

    def items(self):
        return [(key, self[key]) for key in self.keys()]

    def values(self):
        return [self[key] for key in self.keys()]

    def clear(self):
        self.counters = []
        self.groups = []
        self.histograms = []
        self.current_group = None
        self.current_histogram = None

    def set_group(self, name, value):
        try:
            details = self.group_map[name]
        except KeyError:
            raise NetstatStatisticsError('Invalid %s group: %s' % (self.name, name))

        group = Group(
            self,
            details['name'],
            value,
            details['counter_map'],
            details.get('subgroups', {}),
            details.get('histogram_map', {}),
        )
        self.groups.append(group)
        self.current_group = group
        return group

    def set_histogram(self, name):
        try:
            details = self.histogram_map[name]
        except KeyError:
            raise NetstatStatisticsError('Invalid %s histogram: %s' % (self.name, name))

        histogram = Histogram(
            self,
            self,
            details['group'],
            details['name'],
            counter_map=details['counter_map']
        )
        self.histograms.append(histogram)
        self.current_histogram = histogram
        return histogram

    def set_direct_counter(self, name, value, counter_map):
        try:
            details = self.counter_map[name]
            counter = Counter(self, self, details['group'], details['name'], value, counter_map=counter_map)
            self.current_group = None
            self.current_histogram = None
            return counter

        except KeyError:
            return None

    def add_counter(self, counter):
        if counter.group not in self:
            self[counter.group] = []
        self[counter.group].append(counter)

    def add(self, line):
        def split_details(data):
            config = {
                'name': data['name'],
                'value': data['value'],
            }
            if 'name_continued' in data:
                config['name'] += ' %s' % data['name_continued']

            details = {}
            for key in data.keys():
                if key not in ('name', 'name_continued', 'value', ):
                    details[key] = data[key]

            return config, details

        def parse_counter(name, value, counter_map={}):
            if self.current_group is not None:
                counter = self.current_group.add_counter(name, value)
                if counter:
                    return counter

            if self.current_histogram is not None:
                counter = self.current_histogram.add_counter(name, value)
                if counter:
                    return counter

            if self.counter_map is not None:
                counter = self.set_direct_counter(name, value, counter_map)
                if counter is not None:
                    self.add_counter(counter)
                    return counter

            if self.group_map is not None and name in self.group_map:
                counter = self.set_group(name, value)
                return counter

            if self.histogram_map is not None and name in self.histogram_map:
                return self.set_histogram(name, value)

            return Counter(self, self, None, name, value, counter_map)

        line = line.strip()
        counter = None
        for regexp in COUNTER_REGEXPS:
            m = regexp.match(line)
            if not m:
                continue
            config, details = split_details(m.groupdict())
            counter = parse_counter(**config)
            break

        if counter is not None:
            return counter

        elif self.histogram_map is not None and line in self.histogram_map:
            return self.set_histogram(line)

        elif self.current_group is not None:
            if line in self.current_group.histogram_map:
                self.current_group.set_histogram(line)


    def as_dict(self):
        data = {
            'name': self.name,
            'counters': {}
        }

        for group in self.keys():
            data['counters'][group] = [counter.as_dict() for counter in self[group]]

        for attr in ( 'groups', 'histograms', ):
            if not getattr(self, attr):
                continue
            data[attr] = [entry.as_dict() for entry in getattr(self, attr)]

        return data


class NetstatStatistics(dict):
    """Parse netstat -s output

    Parser for netstat -s output. Contains Group, Histogram and Counter objects.
    """
    def __init__(self):
        if sys.platform in ( 'linux2', ):
            self.__lineparser_dict__ = NETSTAT_PROTOCOL_COUNTERS_LINUX

        elif fnmatch.fnmatch(sys.platform, '*bsd*') or sys.platform in ( 'darwin', ):
            self.__lineparser_dict__ = NETSTAT_PROTOCOL_COUNTERS_BSD

        else:
            raise NotImplementedError('Support for %s not yet implemented.' % sys.platform)

        for name in sorted(self.__lineparser_dict__.keys()):
            self[name] = Protocol(name, **self.__lineparser_dict__[name])
        self.updated = datetime.now()

    @property
    def protocols(self):
        return self.values()

    @property
    def all_counters(self):
        counters = []

        for protocol in self.values():
            counters.extend(protocol.all_counters)

        return counters

    @property
    def delta(self):
        ns = NetstatStatistics()
        ns.load()

        counters = []
        for counter in self.all_counters:
            new = ns.find_by_path(counter.path)
            if not new:
                continue
            if counter.value != new.value:
                counters.append(new)

        return ns, counters

    def load(self, protocols=[]):
        p = Popen(['netstat', '-s'], stdin=PIPE, stdout=PIPE, stderr=PIPE)
        stdout, stderr = p.communicate()

        for protocol in self.protocols:
            protocol.clear()

        protocol = None
        for line in [line.rstrip() for line in stdout.splitlines()]:
            if RE_PROTOCOL.match(line):
                try:
                    protocol = self[line.lower().rstrip(':')]
                except KeyError:
                    continue

            elif protocol is not None:
                if not protocols or protocol.name in protocols:
                    counter = protocol.add(line)

    def find_by_path(self, path):
        for counter in self.all_counters:
            if counter.path == path:
                return counter
        return None


    def as_dict(self):
        return {
            'host': os.uname()[1],
            'platform': sys.platform,
            'updated': self.updated.strftime(DATE_FORMAT),
            'protocols': [protocol.as_dict() for protocol in self.protocols]
        }


def counter_match_pattern(counter, pattern):
    """Match pattern to counter path

    Returns True if given counter matches the pattern provided. Patterns can be simple
    strings or dot separated path patterns (for example tcp.sent.*).
    """
    if not isinstance(counter, Counter):
        raise NetstatStatisticsError('Not a Counter instance: %s' % counter)

    pattern_parts = [part for part in pattern.split('.') if part != '']
    counter_parts = counter.path.split('.')

    if len(pattern_parts) == 1:
        if fnmatch.fnmatch(counter.path, pattern):
            return True

    for i, match in enumerate(pattern_parts):
        if not fnmatch.fnmatch(counter_parts[i], match):
            return False

    return True


def filter_counters(counters, patterns, negate=False):
    """Filter counters by patterns

    Filters counters with counter_match_pattern matching any of patterns provided.

    If negate is True, patterns NOT matching any of patterns are returned.
    """
    if isinstance(patterns, basestring):
        patterns = patterns.split(',')

    if not patterns:
        return counters

    matched = []
    for pattern in patterns:
        for counter in counters:
            if counter_match_pattern(counter, pattern):
                if not negate:
                    matched.append(counter)
            else:
                if negate:
                    matched.append(counter)

    return matched




"""
Reverse IP delegation parser from ARIN
"""

import requests
import json
import logging

from dateutil.parser import parse as dateparser
from seine.whois import WhoisError
from seine.address import IPv4Address, IPv6Address
from systematic.log import Logger

URL_TEMPLATE = 'http://whois.arin.net/rest/ip/%(address)s'

DATE_FIELD_MAP = {
    'registered': 'registrationDate',
    'updated': 'updateDate',
}

logger = Logger().default_stream

class ARINNetBlock(object):
    """Netblock entry

    Netblock entry in ARINReverseIP results

    """
    def __init__(self, reverse, data):
        self.reverse = reverse
        self.description = data['description']['$']
        self.type = data['type']['$']
        self.mask = int(data['cidrLength']['$'])

        if reverse.address_format == IPv4Address:
            self.start = IPv4Address(data['startAddress']['$'])
            self.end = IPv4Address(data['endAddress']['$'])
            self.network = IPv4Address('%s/%s' % (self.start.address, self.mask))

        elif reverse.address_format == IPv6Address:
            self.start = IPv6Address(data['startAddress']['$'])
            self.end = IPv6Address(data['endAddress']['$'])
            self.network = IPv6Address('%s/%s' % (self.start.address, self.mask))

    def __repr__(self):
        return '%s' % self.network

    def match(self, address):
        try:
            if self.network.addressInNetwork(IPv4Address(address).address):
                return self
        except ValueError:
            try:
                if self.network.addressInNetwork(IPv6Address(address).address):
                    return self
            except ValueError:
                raise WhoisError('Error matching address %s to %s' % (address, self.network))

        return None

class ARINReverseIP(list):
    """ARIN reverse query response

    List of networks parsed from ARIN reverse IP query

    """
    def __init__(self, address):
        try:
            self.address = IPv4Address(address).ipaddress
            self.address_format = IPv4Address
        except ValueError:
            try:
                self.address = IPv6Address(address).address
                self.address_format = IPv6Address
            except ValueError:
                raise WhoisError('Unsupported address: %s' % address)

    def __hash__(self):
        return sum([x.network.raw_value for x in self])

    def __repr__(self):
        return '%s ARIN response %d netblocks' % (self.address, len(self))

    def __parse_date_entry(self, data):
        if data is None:
            return None
        try:
            return dateparser(data['$'])
        except ValueError, KeyError:
            raise WhoisError('Error parsing date from %s' % data)

    def __parse_number_entry(self, data):
        if data is None:
            return None
        try:
            return int(data['$'])
        except KeyError:
            raise WhoisError('Error parsing number field %s' % data)

    def __parse_string_entry(self, data):
        if data is None:
            return None
        if 'line' in data:
            try:
                data = data['line']
                if isinstance(data, list):
                    data.sort(lambda a, b: cmp(int(a['@number']), int(b['@number'])))
                    return ' '.join([l['$'] for l in data])
                else:
                    return data['$']
            except AttributeError:
                raise WhoisError('Error parsing string field %s' % data)

        else:
            try:
                return data['$']
            except TypeError:
                raise WhoisError('Error parsing string field %s' % data)
            except KeyError:
                raise WhoisError('Error parsing string field %s' % data)

    def __parse_address_entry(self, data):
        if data is None:
            return None
        try:
            return self.address_format(data['$'])
        except ValueError, KeyError:
            raise WhoisError('Error parsing address from %s' % data)

    @property
    def name(self):
        return hasattr(self, '_name') and self._name or ''
    @name.setter
    def name(self, value):
        self._name = self.__parse_string_entry(value)

    @property
    def handle(self):
        return hasattr(self, '_handle') and self._handle or ''
    @handle.setter
    def handle(self, value):
        self._handle = self.__parse_string_entry(value)

    @property
    def ref(self):
        return hasattr(self, '_ref') and self._ref or ''
    @ref.setter
    def ref(self, value):
        self._ref = self.__parse_string_entry(value)

    @property
    def comment(self):
        return hasattr(self, '_comment') and self._comment or ''
    @comment.setter
    def comment(self, value):
           self._comment = self.__parse_string_entry(value)

    @property
    def version(self):
        return hasattr(self, '_version') and self._version or 0
    @version.setter
    def version(self, value):
        self._version = self.__parse_number_entry(value)

    @property
    def registered(self):
        return hasattr(self, '_registered') and self._registered or None
    @registered.setter
    def registered(self, value):
        self._registered = self.__parse_date_entry(value)

    @property
    def updated(self):
        return hasattr(self, '_updated') and self._updated or None
    @updated.setter
    def updated(self, value):
        self._updated = self.__parse_date_entry(value)

    def keys(self):
        return ['name', 'handle', 'ref', 'registered', 'updated', 'version', 'comment',]

    def items(self):
        return [(key, getattr(self, key)) for key in self.keys()]

    def values(self):
        return [getattr(self, key) for key in self.keys()]

    def match(self, address):
        """Check if address matches

        Return netblock if address matches any of our networks, otherwise None

        Exceptions matching netblock are silently ignored.
        """
        for netblock in self:
            try:
                if netblock.match(address):
                    return netblock
            except WhoisError:
                continue

        return None

def ARINReverseIPQuery(address, proxies={}):
    """ARIN IP Query

    Query ARIN REST API for IP allocation information details

    Returns ARINReverseIP object.

    """
    entry = ARINReverseIP(address)

    url = URL_TEMPLATE % {'address': entry.address}
    headers = {'Accept': 'application/arin.whoisrws-v1+json' }
    res = requests.get(url, headers=headers)
    if res.status_code != 200:
        raise WhoisError('Error fetching URL %s' % url)

    try:
        data = json.loads(res.content)
    except ValueError, emsg:
        raise WhoisError('Error parsing response: %s' % res.content)

    if 'net' not in data:
        logger.debug('DATA: %s' % res.content)
        raise WhoisError('Did not receive expected data: missing net section')
    net = data['net']
    assert isinstance(net, dict)

    for key in ( 'handle', 'name', 'comment', 'ref', 'version', ):
        if key in net:
            setattr(entry, key, net[key])

    for key, field in DATE_FIELD_MAP.items():
        try:
            setattr(entry, key, net[field])
        except KeyError:
            continue

    blocks = net['netBlocks']['netBlock']
    if isinstance(blocks, list):
        entry.extend(ARINNetBlock(entry, block) for block in blocks)
    else:
        entry.append(ARINNetBlock(entry, blocks))

    return entry




"""
Module to update and process the list of valid TLD domains from IANA.
"""

import os
import re

from seine.url import HTTPRequest, HTTPRequestError

IANA_TLD_URL = 'http://data.iana.org/TLD/tlds-alpha-by-domain.txt'

CACHE_DIRECTORIES = [
    '/var/cache/tld',
    '/tmp/tld-%s' % os.geteuid(),
]

CACHE_FILES = map(lambda d:
    os.path.join(d, os.path.basename(IANA_TLD_URL)),
    CACHE_DIRECTORIES
)

RE_TLDNAME = re.compile(r'^[a-z0-9-]+$')


class DNSCacheError(Exception):
    pass


class TLD(object):
    def __init__(self, name):
        self.name = name

    def __repr__(self):
        if self.unicode != self.name:
            return '%s (%s)' % (self.name, self.unicode)
        else:
            return self.name

    def __cmp__(self, other):
        if isinstance(other, TLD):
            return cmp(self.unicode, other.unicode)

        if other == self.name or other == self.unicode:
            return 0

        return cmp(self.unicode, other)

    @property
    def unicode(self):
        return unicode(self.name).decode('idna')

    @property
    def is_idn(self):
        return self.name != self.unicode


class TLDCache(list):
    """
    Abstratction for TLD domain names: you can
    - update the list from IANA with updateCache() method
    - check if a given value is valid TLD with has_key
    - iterate all valid TLDs by iteraing the class itself
    """
    def __init__(self, path=None):
        """
        You can pass alternative file path to be used in all operations,
        or use the default file.
        """

        self.__next = 0
        self.path = path

        if self.path is None:
            for f in CACHE_FILES:
                if os.path.isfile(f) and os.access(f, os.W_OK):
                    self.path = f
                    break

                fdir = os.path.dirname(f)
                if not os.path.isdir(fdir):
                    try:
                        os.makedirs(os.path.dirname(f))
                    except IOError, (ecode, emsg):
                        continue
                    except OSError, (ecode, emsg):
                        continue

                if not os.path.isfile(f):
                    try:
                        open(f, 'w').write('\n')
                        os.unlink(f)
                        self.path = f
                        break
                    except IOError, (ecode, emsg):
                        continue
                    except OSError, (ecode, emsg):
                        continue

        self.sort()

    def __repr__(self):
        return 'TLDCache %d entries' % len(self)

    def __getitem__(self, item):
        if len(self) == 0:
            self.update()
        try:
            return list.__getitem__(self, self.index(item))
        except ValueError:
            raise KeyError('No such TLDCache item: %s' % item)

    @property
    def is_downloaded(self):
        return len(self) > 0

    def load(self):
        """Load saved file

        Load saved TLD mapping cache information file

        """

        self.__delslice__(0, len(self))

        if not os.path.isfile(self.path):
            return

        try:
            fd = open(self.path, 'r')
            for l in map(lambda x: x.strip(), fd.readlines()):
                if l.strip() == '' or l.startswith('#'):
                    continue
                if not re.match(RE_TLDNAME, l.lower()):
                    continue
                self.append(TLD(l.lower()))

        except IOError, (ecode, emsg):
            raise TLDCacheError('Error loading cache: %s' % emsg)

        except OSError, (ecode, emsg):
            raise TLDCacheError('Error loading cache: %s' % emsg)

    def update(self):
        """Update cache

        Compatibility callback for self.download()

        """
        return self.download()

    def download(self):
        """Update TLD cache file

        Download a new TLD cache file from IANA and load new version
        with self.load()

        """

        cache_dir = os.path.dirname(self.path)

        if not os.path.isdir(cache_dir):
            try:
                os.mkdir(cache_dir)
            except OSError, (ecode, emsg):
                raise DNSCacheError('Error creating %s: %s' % (cache_dir, emsg))

        try:
            req = HTTPRequest()
            (code, data, headers) = req.GET(IANA_TLD_URL)
            fd = open(self.path, 'w')
            fd.write(data)
            fd.close()

        except IOError, emsg:
            raise DNSCacheError('Error updating %s: %s' % (self.path, emsg))

        except OSError, (ecode, emsg):
            raise DNSCacheError('Error updating %s: %s' % (self.path, emsg))

        self.load()




#!/usr/bin/env python
"""
Parser for ficora whois data formats
"""

import re
from datetime import datetime
from seine.whois.formats import WhoisDataParser, WhoisError

FIELD_MAP = {
    'domain':           'domainname',
    'descr':            'owner',
    'address':          'address',
    'phone':            'telephone',
    'status':           'status',
    'nserver':          'nameservers',
    'created':          'created',
    'modified':         'updated',
    'expires':          'expires',
    'dnssec':           'dnssec',
}


RE_NS_STATUS = re.compile('(?P<ns>.*) \[(?P<status>.*)\]$')

DNS_FIELDS = ( 'nameservers', 'whois_servers', 'domainname', )
DATE_FIELDS = ( 'updated', 'created', 'expires', )
DATE_PARSER = lambda value: datetime.strptime(value, '%d.%m.%Y').date()

class ficora(WhoisDataParser):
    tlds = ( 'fi', )

    def parse(self, domain, data):
        data = WhoisDataParser.parse(self, domain, data)

        for l in data:
            if l.count(': ') == 0:
                continue
            try:
                k, v = [x.strip() for x in l.split(':', 1)]
            except ValueError:
                raise WhoisError('Error parsing line %s' % l)

            try:
                k = FIELD_MAP[k]
            except KeyError:
                raise WhoisError('Unknown field on line %s' % l)

            if k in DATE_FIELDS:
                v = DATE_PARSER(v)

            elif k == 'nameservers':
                m = RE_NS_STATUS.match(v)
                if not m:
                    raise WhoisError('Error parsing line %s' % l)
                v = m.groupdict()['ns']

            elif k in DNS_FIELDS:
                v = v.lower()

            elif k == 'dnssec':
                v = v != 'no' and True or False

            self.set(k, v)





import os
import re
import requests

from seine.address import EthernetMACAddress

OUI_URL = 'http://standards.ieee.org/develop/regauth/oui/oui.txt'
OUI_CACHE_PATH = os.path.expanduser('~/.oiu.cache')

HEX_LINE = re.compile('\s+(?P<prefix>[A-Z0-9-]+)\s+\(hex\)\s+(?P<company>.*)$')
BASE_16_LINE = re.compile('\s+(?P<prefix>[A-Z0-9]+)\s+\(base 16\)\s+(?P<company>.*)$')


class OUIPrefix(object):
    def __init__(self, prefix, company):
        self.prefix = ':'.join(prefix.split('-')).upper()
        self.company = company
        self.address = []

    def __repr__(self):
        return '%s %s' % (self.company, ', '.join(self.address))


class OUIPrefixLookup(dict):
    def __init__(self, path=OUI_CACHE_PATH):
        self.path = path
        if not os.path.isfile(self.path):
            self.update()

        entry = None
        for l in [x.rstrip() for x in open(self.path, 'r').readlines()]:
            if l.strip() == '' or  BASE_16_LINE.match(l):
                continue

            m = HEX_LINE.match(l)
            if m:
                entry = OUIPrefix(**m.groupdict())
                self[entry.prefix] = entry
                continue

            elif entry:
                entry.address.append(l.strip())

    def update(self):
        res = requests.get(OUI_URL)
        open(self.path, 'w').write(res.content)

    def match(self, address):
        if not isinstance(address, EthernetMACAddress):
            address = EthernetMACAddress(address)

        parts = address.address.upper().split(':')
        while parts:
            key = ':'.join(parts)
            if key in self:
                return self[key]
            parts.pop()

        return None




#!/usr/bin/env python
"""
Parse and fetch latest root servers database file.
"""

import os
import random
import requests

from seine.dns import DNSError
from seine.address import IPv4Address, IPv6Address

UPDATE_URLS = [
    'http://ftp.internic.net/domain/named.root',
    'htp://rs.internic.net/domain/named.root',
]


class RootServerAddressEntry(object):
    def __init__(self, server, address, expire):
        self.server = server
        if not isinstance(address, IPv4Address) and not isinstance(address, IPv6Address):
            raise ValueError('Unsupported address type: %s' % type(address))

        self.address_record = address
        self.expire = expire

    def __repr__(self):
        return self.address

    @property
    def address(self):
        if isinstance(self.address_record, IPv4Address):
            return self.address_record.ipaddress

        elif isinstance(self.address_record, IPv6Address):
            return self.address_record.address

class RootServer(object):
    """
    Represents one root server DNS name. Contains lists of IPv4 and
    IPv6 addresses with cache expiry times for each address associated
    to this server. The IPv4 and IPv6 servers can be accessed by the
    'A' and 'AAAA' keys.
    """
    def __init__(self, name, expire=0):
        self.name = name.lower()
        self.expire = expire
        self.ipv4_addresses = []
        self.ipv6_addresses = []

    def __repr__(self):
        return self.name

    def add(self, rrname, address, expire):
        """
        Adds a A or AAAA RR, associated address and cache expiry value
        to the server.
        """
        rrname = rrname.upper()
        if rrname == 'A':
            try:
                self.ipv4_addresses.append(RootServerAddressEntry(self, IPv4Address(address), expire))
            except ValueError:
                raise DNSError('Invalid IPv4 address: %s' % address)

        elif rrname == 'AAAA':
            try:
                self.ipv6_addresses.append(RootServerAddressEntry(self, IPv6Address('%s/128' % address), expire))
            except ValueError:
                raise DNSError('Invalid IPv6 address: %s' % address)

        else:
            raise DNSError('Unsupported root server address RR: %s' % rrname)


class RootServers(dict):
    """
    Loads, parses and updates the named.root offcial root server data
    file (if requested). Provides methods to list both IPv4 and IPv6
    root servers, by name or address.
    """
    def __init__(self, path):
        self.path = os.path.expandvars(os.path.expanduser(path))

        if os.path.isfile(self.path):
            self.parse()

    def __download_data__(self, update_urls=UPDATE_URLS):
        for url in update_urls:
            res = requests.get(url)
            if res.status_code != 200:
                continue

            return res.content

        raise DNSError('Error downloading root server list')

    def download(self, update_urls=UPDATE_URLS):
        """
        Downloads and replaces the named.root file from official sources.

        Replaces current status on the fly if download was successful.

        """
        data = self.__download_data__(update_urls)

        try:
            open(self.path, 'w').write(data)

        except IOError, (ecode, emsg):
            raise DNSError('Error writing %s: %s' % (self.path, emsg))
        except IOError, (ecode, emsg):
            raise DNSError('Error writing %s: %s' % (self.path, emsg))

        try:
            self.parse()

        except DNSError, emsg:
            raise DNSError('Update of root server list from %s failed: %s' % (self.path, emsg))

    def parse(self):
        """
        Load the root servers list from the file given when initializing.
        """
        self.clear()
        if not os.path.isfile(self.path):
            raise DNSError('No such file: %s' % self.path)

        for line in [l.strip() for l in open(self.path, 'r').readlines()]:
            if line[:1] in ( ';'):
                continue

            if line[:1] == '.':
                try:
                    (expire, rrclass, rrname, name) = line[1:].split(None, 3)
                except ValueError:
                    (expire, rrname, name) = line[1:].split(None, 2)

                if rrname != 'NS':
                    raise DNSError('Error parsing root servers file line: %s' % line)

                if name not in self:
                    self[name] = RootServer(name, expire)

            else:
                try:
                    (name, expire, rrname, address) = line.split()
                except ValueError:
                    raise DNSError('Error parsing root servers file line: %s' % line)

                if name not in self.keys():
                    raise DNSError('Unexpected server line: %s' % line)

                self[name].add(rrname, address, expire)

    @property
    def is_downloaded(self):
        return os.path.isfile(self.path)

    @property
    def ipv4_server_names(self):
        """
        Return names of IPv4 root servers (with A record)
        """
        return [k for k in self.keys() if self[k].ipv4_addresses]

    @property
    def ipv6_server_names(self):
        """
        Return names of IPv6 root servers (with AAAA record)
        """
        return [k for k in self.keys() if self[k].ipv6_addresses]

    @property
    def ipv4_server_addresses(self):
        """
        Return addresses of IPv4 root servers (with A record)
        """
        addresses = []
        for server in [s for s in self.values() if s.ipv4_addresses]:
            for addr in server.ipv4_addresses:
                if addr not in addresses:
                    addresses.append(addr)
        return addresses

    @property
    def ipv6_server_addresses(self):
        """
        Return addresses of IPv6 root servers (with AAAA record)
        """
        addresses = []
        for server in [s for s in self.values() if s.ipv6_addresses]:
            for addr in server.ipv6_addresses:
                if addr not in addresses:
                    addresses.append(addr)
        return addresses

    def random_rootserver(self, addresstype=IPv4Address):
        """Return random root server

        Optionally specify addresstype as ipv4 or ipv6 to return address
        of this type. Default is IPv4.

        """
        if addresstype in ('A', 'ipv4', IPv4Address):
            servers = self.ipv4_server_addresses
            addresstype = 'ipv4'
        elif addresstype in ('AAAA', 'ipv6', IPv6Address):
            servers = self.ipv6_server_addresses
            addresstype = 'ipv6'
        else:
            raise DNSError('Invalid addresstype argument: %s' % addresstype)

        if not servers:
            raise DNSError('Error choosing random server: no root servers matching %s available' & addresstype)
        return random.choice(servers)


#!/usr/bin/env python

import os
import glob
from setuptools import setup, find_packages
from seine import __version__

setup(
    name = 'seine',
    version = __version__,
    license = 'PSF',
    keywords = 'network utility scripts and classes for python',
    url = 'https://github.com/hile/seine',
    author = 'Ilkka Tuohela',
    author_email = 'hile@iki.fi',
    description = 'Various network address and url related utilities',
    packages = find_packages(),
    scripts = glob.glob('bin/*'),
    install_requires = (
        'systematic>=4.4.0',
        'requests',
        'python-dateutil',
        'dnspython',
        'pyip',
        'pyasn1',
        'pysnmp'
    ),
)

